==> workflow_graph-0.1.0.tar.gz: workflow_graph-0.1.0/README.md <==
# WorkflowGraph

**WorkflowGraph** is a lightweight, self-contained Python library for building and executing directed graph workflows. It's an alternative to **LangGraph** for those seeking independence from LangChain and the flexibility to implement agent workflows, while still enabling real-time streaming of results.

The JavaScript-version is available at [https://github.com/dextersjab/workflow-graph.js]().

## Features

- **Graph-based workflows**: Build flexible, directed workflows where nodes are customizable tasks.
- **Synchronous & asynchronous support**: Define both sync and async nodes without any external dependencies.
- **Real-time streaming**: Built-in support for callbacks in each node, allowing real-time token streaming (e.g., for WebSockets).
- **LangGraph alternative**: Unlike LangGraph, WorkflowGraph provides a simpler, fully self-contained solution without needing LangChain for streaming.
- **Modular architecture**: Organized into separate modules for better maintainability and extensibility.

## Installation

To install **WorkflowGraph** as a dependency, add the following line to your `requirements.txt`:

```
git+https://github.com/dextersjab/workflow-graph.git@main
```

Then, run:

```shell
pip install -r requirements.txt
```

Or install directly using pip:

```shell
pip install git+https://github.com/dextersjab/workflow-graph.git@main
```

## Basic Usage

### Implementing a Workflow

Here's how to create a simple workflow with conditional branching:

```python
import asyncio
from workflow_graph import WorkflowGraph

# Define task functions
def add(data, callback=None):
    # The optional callback parameter can be used for streaming interim results,
    # logging progress, or updating UI in real-time as the workflow executes
    result = data + 1
    if callback:
        callback(f"Added 1: {data} -> {result}")
    return result

def is_even(data):
    return data % 2 == 0

def handle_even(data):
    return f"Even: {data}"

def handle_odd(data):
    return f"Odd: {data}"

# Create and configure the workflow graph
graph = WorkflowGraph()

# Add nodes
graph.add_node("addition", add)
graph.add_node("is_even_check", is_even)
graph.add_node("even_handler", handle_even)
graph.add_node("odd_handler", handle_odd)

# Define starting point
graph.set_entry_point("addition")

# Define flow between nodes
graph.add_edge("addition", "is_even_check")

# Add conditional branching based on is_even_check result
graph.add_conditional_edges(
    "is_even_check", 
    path=is_even, 
    path_map={True: "even_handler", False: "odd_handler"}
)

# Set endpoints
graph.set_finish_point("even_handler")
graph.set_finish_point("odd_handler")
```

This example creates a workflow that:
1. Takes a number as input
2. Adds 1 to it
3. Checks if the result is even
4. Branches to different handlers based on the result

```mermaid
flowchart TD
    __start__["START"]
    __end__["END"]
    addition["addition"]
    is_even_check["is_even_check"]
    even_handler["even_handler"]
    odd_handler["odd_handler"]
    __start__ --> addition
    addition --> is_even_check
    is_even_check -.|True|.-> even_handler
    is_even_check -.|False|.-> odd_handler
    even_handler --> __end__
    odd_handler --> __end__
```

### Error Handling and Retries

WorkflowGraph supports built-in error handling and retry capabilities:

```python
graph.add_node(
    "api_call", 
    make_api_request, 
    retries=3,                   # Retry up to 3 times on failure
    backoff_factor=0.5,          # Wait 0.5 seconds × attempt before retrying
    on_error=handle_api_error    # Call this function if all retries fail
)
```

## Execution Methods

Once your workflow is defined, there are two ways to execute it:

### Direct Execution (Simple)

For one-time executions, use the direct execution approach:

```python
# Execute synchronously
result = graph.execute(input_data)

# Or execute asynchronously with a callback
result = await graph.execute_async(input_data, callback=some_callback)
```

### Compile-then-Execute (More Efficient for Multiple Executions)

For workflows that will be executed multiple times, compile once and reuse:

```python
# Compile the graph
compiled_graph = graph.compile()

async def run_workflow(input_data):
    # Execute with the compiled graph
    result = await compiled_graph.execute_async(
        input_data, 
        callback=lambda msg: print(f"Progress update: {msg}")
    )
    print(f"Final Result: {result}")

# Run the workflow with different inputs
asyncio.run(run_workflow(5))
asyncio.run(run_workflow(10))
```

## Generating Mermaid Diagrams

WorkflowGraph includes built-in support for generating [Mermaid.js](https://mermaid.js.org/) diagrams to visualize your workflow:

```python
# Generate Mermaid diagram code
mermaid_code = graph.to_mermaid()
print(mermaid_code)
```

The generated diagram uses dashed lines (`-.->`), rather than decision nodes, to represent conditional branches. This provides a cleaner and more accurate representation of how the workflow behaves.

Mermaid diagrams can be rendered in:
- GitHub Markdown (just paste the code)
- VS Code (with the Mermaid extension)
- Web browsers (using the Mermaid Live Editor)
- Many other tools that support Mermaid

## Package Structure

The library is organized into the following modules:

- **workflow_graph**: Main package
- **constants.py**: Defines constants like START and END
- **models.py**: Defines data structures like NodeSpec and Branch
- **builder.py**: Contains the WorkflowGraph class for building graphs
- **executor.py**: Contains the CompiledGraph class for executing workflows
- **exceptions.py**: Contains custom exceptions for better error handling
- **README.md**: Contains detailed documentation about the package internals

For backward compatibility, a top-level `workflow_graph.py` file is also provided that re-exports all the public API.

> **Note**:
> - For more technical details about the package internals, refer to the [src/workflow_graph/README.md](src/workflow_graph/README.md) file.
> - This project is largely AI-generated

==> workflow_graph-0.1.1.tar.gz: workflow_graph-0.1.1/README.md <==
# WorkflowGraph

**WorkflowGraph** is a lightweight, self-contained Python library for building and executing directed graph workflows. It's an alternative to **LangGraph** for those seeking independence from LangChain and the flexibility to implement agent workflows, while still enabling real-time streaming of results.

The JavaScript-version is available at [https://github.com/dextersjab/workflow-graph.js]().

## Features

- **Graph-based workflows**: Build flexible, directed workflows where nodes are customizable tasks.
- **Synchronous & asynchronous support**: Define both sync and async nodes without any external dependencies.
- **Real-time streaming**: Built-in support for callbacks in each node, allowing real-time token streaming (e.g., for WebSockets).
- **LangGraph alternative**: Unlike LangGraph, WorkflowGraph provides a simpler, fully self-contained solution without needing LangChain for streaming.
- **Modular architecture**: Organized into separate modules for better maintainability and extensibility.

## Installation

### From PyPI (Recommended)

```shell
pip install workflow-graph
```

### From GitHub

Add the following line to your `requirements.txt`:

```
git+https://github.com/dextersjab/workflow-graph.git@main
```

Then, run:

```shell
pip install -r requirements.txt
```

Or install directly using pip:

```shell
pip install git+https://github.com/dextersjab/workflow-graph.git@main
```

## Basic Usage

### Implementing a Workflow

Here's how to create a simple workflow with conditional branching:

```python
import asyncio
from workflow_graph import WorkflowGraph

# Define task functions
def add(data, callback=None):
    # The optional callback parameter can be used for streaming interim results,
    # logging progress, or updating UI in real-time as the workflow executes
    result = data + 1
    if callback:
        callback(f"Added 1: {data} -> {result}")
    return result

def is_even(data):
    return data % 2 == 0

def handle_even(data):
    return f"Even: {data}"

def handle_odd(data):
    return f"Odd: {data}"

# Create and configure the workflow graph
graph = WorkflowGraph()

# Add nodes
graph.add_node("addition", add)
graph.add_node("is_even_check", is_even)
graph.add_node("even_handler", handle_even)
graph.add_node("odd_handler", handle_odd)

# Define starting point
graph.set_entry_point("addition")

# Define flow between nodes
graph.add_edge("addition", "is_even_check")

# Add conditional branching based on is_even_check result
graph.add_conditional_edges(
    "is_even_check", 
    path=is_even, 
    path_map={True: "even_handler", False: "odd_handler"}
)

# Set endpoints
graph.set_finish_point("even_handler")
graph.set_finish_point("odd_handler")
```

This example creates a workflow that:
1. Takes a number as input
2. Adds 1 to it
3. Checks if the result is even
4. Branches to different handlers based on the result

```mermaid
flowchart TD
    __start__["START"]
    __end__["END"]
    addition["addition"]
    is_even_check["is_even_check"]
    even_handler["even_handler"]
    odd_handler["odd_handler"]
    __start__ --> addition
    addition --> is_even_check
    is_even_check -.|True|.-> even_handler
    is_even_check -.|False|.-> odd_handler
    even_handler --> __end__
    odd_handler --> __end__
```

### Error Handling and Retries

WorkflowGraph supports built-in error handling and retry capabilities:

```python
graph.add_node(
    "api_call", 
    make_api_request, 
    retries=3,                   # Retry up to 3 times on failure
    backoff_factor=0.5,          # Wait 0.5 seconds × attempt before retrying
    on_error=handle_api_error    # Call this function if all retries fail
)
```

## Execution Methods

Once your workflow is defined, there are two ways to execute it:

### Direct Execution (Simple)

For one-time executions, use the direct execution approach:

```python
# Execute synchronously
result = graph.execute(input_data)

# Or execute asynchronously with a callback
result = await graph.execute_async(input_data, callback=some_callback)
```

### Compile-then-Execute (More Efficient for Multiple Executions)

For workflows that will be executed multiple times, compile once and reuse:

```python
# Compile the graph
compiled_graph = graph.compile()

async def run_workflow(input_data):
    # Execute with the compiled graph
    result = await compiled_graph.execute_async(
        input_data, 
        callback=lambda msg: print(f"Progress update: {msg}")
    )
    print(f"Final Result: {result}")

# Run the workflow with different inputs
asyncio.run(run_workflow(5))
asyncio.run(run_workflow(10))
```

## Generating Mermaid Diagrams

WorkflowGraph includes built-in support for generating [Mermaid.js](https://mermaid.js.org/) diagrams to visualize your workflow:

```python
# Generate Mermaid diagram code
mermaid_code = graph.to_mermaid()
print(mermaid_code)
```

The generated diagram uses dashed lines (`-.->`), rather than decision nodes, to represent conditional branches. This provides a cleaner and more accurate representation of how the workflow behaves.

Mermaid diagrams can be rendered in:
- GitHub Markdown (just paste the code)
- VS Code (with the Mermaid extension)
- Web browsers (using the Mermaid Live Editor)
- Many other tools that support Mermaid

## Package Structure

The library is organized into the following modules:

- **workflow_graph**: Main package
- **constants.py**: Defines constants like START and END
- **models.py**: Defines data structures like NodeSpec and Branch
- **builder.py**: Contains the WorkflowGraph class for building graphs
- **executor.py**: Contains the CompiledGraph class for executing workflows
- **exceptions.py**: Contains custom exceptions for better error handling
- **README.md**: Contains detailed documentation about the package internals

For backward compatibility, a top-level `workflow_graph.py` file is also provided that re-exports all the public API.

> **Note**:
> - For more technical details about the package internals, refer to the [src/workflow_graph/README.md](src/workflow_graph/README.md) file.
> - This project is largely AI-generated
